# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' keystone/exception.py
KEYSTONE_EXCEPTIONS = [
    "_KeystoneExceptionMeta",
    "Error",
    "ValidationError",
    "URLValidationError",
    "PasswordValidationError",
    "PasswordRequirementsValidationError",
    "PasswordHistoryValidationError",
    "PasswordAgeValidationError",
    "PasswordSelfServiceDisabled",
    "SchemaValidationError",
    "ValidationTimeStampError",
    "InvalidOperatorError",
    "ValidationExpirationError",
    "StringLengthExceeded",
    "AmbiguityError",
    "ApplicationCredentialValidationError",
    "CircularRegionHierarchyError",
    "ForbiddenNotSecurity",
    "PasswordVerificationError",
    "RegionDeletionError",
    "ApplicationCredentialLimitExceeded",
    "CredentialLimitExceeded",
    "SecurityError",
    "Unauthorized",
    "InsufficientAuthMethods",
    "ReceiptNotFound",
    "PasswordExpired",
    "AuthPluginException",
    "UserDisabled",
    "AccountLocked",
    "AuthMethodNotSupported",
    "ApplicationCredentialAuthError",
    "AdditionalAuthRequired",
    "Forbidden",
    "ForbiddenAction",
    "CrossBackendNotAllowed",
    "InvalidPolicyAssociation",
    "InvalidDomainConfig",
    "InvalidLimit",
    "LimitTreeExceedError",
    "NotFound",
    "EndpointNotFound",
    "PolicyNotFound",
    "PolicyAssociationNotFound",
    "RoleNotFound",
    "ImpliedRoleNotFound",
    "InvalidImpliedRole",
    "DomainSpecificRoleMismatch",
    "DomainSpecificRoleNotWithinIdPDomain",
    "DomainIdInvalid",
    "RoleAssignmentNotFound",
    "RegionNotFound",
    "ServiceNotFound",
    "DomainNotFound",
    "ProjectNotFound",
    "ProjectTagNotFound",
    "TokenNotFound",
    "UserNotFound",
    "GroupNotFound",
    "MappingNotFound",
    "TrustNotFound",
    "TrustUseLimitReached",
    "CredentialNotFound",
    "VersionNotFound",
    "EndpointGroupNotFound",
    "IdentityProviderNotFound",
    "ServiceProviderNotFound",
    "FederatedProtocolNotFound",
    "PublicIDNotFound",
    "RegisteredLimitNotFound",
    "LimitNotFound",
    "NoLimitReference",
    "RegisteredLimitError",
    "DomainConfigNotFound",
    "ConfigRegistrationNotFound",
    "ApplicationCredentialNotFound",
    "AccessRuleNotFound",
    "Conflict",
    "UnexpectedError",
    "TrustConsumeMaximumAttempt",
    "MalformedEndpoint",
    "MappedGroupNotFound",
    "MetadataFileError",
    "DirectMappingError",
    "AssignmentTypeCalculationError",
    "NotImplemented",
    "Gone",
    "ConfigFileNotFound",
    "KeysNotFound",
    "MultipleSQLDriversInConfig",
    "MigrationNotProvided",
    "UnsupportedTokenVersionException",
    "SAMLSigningError",
    "OAuthHeadersMissingError",
    "TokenlessAuthConfigError",
    "CredentialEncryptionError",
    "LDAPServerConnectionError",
    "LDAPInvalidCredentialsError",
    "LDAPSizeLimitExceeded",
    "CacheDeserializationError",
    "ResourceUpdateForbidden",
    "ResourceDeleteForbidden",
]
