
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' barbican/common/exception.py
BARBICAN_EXCEPTIONS = [
    "BarbicanException",
    "BarbicanHTTPException",
    "MissingArgumentError",
    "MissingMetadataField",
    "InvalidMetadataRequest",
    "InvalidMetadataKey",
    "InvalidSubjectDN",
    "InvalidContainer",
    "InvalidExtensionsData",
    "InvalidCMCData",
    "InvalidPKCS10Data",
    "InvalidCertificateRequestType",
    "CertificateExtensionsNotSupported",
    "FullCMCNotSupported",
    "NotFound",
    "ConstraintCheck",
    "NotSupported",
    "Invalid",
    "NoDataToProcess",
    "LimitExceeded",
    "InvalidObject",
    "PayloadDecodingError",
    "UnsupportedField",
    "FeatureNotImplemented",
    "StoredKeyContainerNotFound",
    "StoredKeyPrivateKeyNotFound",
    "ProvidedTransportKeyNotFound",
    "InvalidCAID",
    "CANotDefinedForProject",
    "QuotaReached",
    "InvalidParentCA",
    "SubCAsNotSupported",
    "SubCANotCreated",
    "CannotDeleteBaseCA",
    "UnauthorizedSubCA",
    "CannotDeletePreferredCA",
    "BadSubCACreationRequest",
    "SubCACreationErrors",
    "SubCADeletionErrors",
    "PKCS11Exception",
    "P11CryptoPluginKeyException",
    "P11CryptoPluginException",
    "P11CryptoKeyHandleException",
    "P11CryptoTokenException",
    "MultipleStorePreferredPluginMissing",
    "MultipleStorePluginStillInUse",
    "MultipleSecretStoreLookupFailed",
    "MultipleStoreIncorrectGlobalDefault",
    "MultipleStorePluginValueMissing",
]
