# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' cinder/exception.py
CINDER_EXCEPTIONS = [
    "ConvertedException",
    "CinderException",
    "VolumeBackendAPIException",
    "VolumeDriverException",
    "BackupDriverException",
    "BackupRestoreCancel",
    "GlanceConnectionFailed",
    "ProgrammingError",
    "NotAuthorized",
    "AdminRequired",
    "PolicyNotAuthorized",
    "ImageNotAuthorized",
    "DriverNotInitialized",
    "Invalid",
    "InvalidSnapshot",
    "InvalidVolumeAttachMode",
    "VolumeAttached",
    "InvalidResults",
    "InvalidInput",
    "InvalidAvailabilityZone",
    "InvalidTypeAvailabilityZones",
    "InvalidVolumeType",
    "InvalidGroupType",
    "InvalidVolume",
    "InvalidContentType",
    "InvalidHost",
    "InvalidParameterValue",
    "InvalidAuthKey",
    "InvalidConfigurationValue",
    "ServiceUnavailable",
    "UnavailableDuringUpgrade",
    "ImageUnacceptable",
    "ImageTooBig",
    "DeviceUnavailable",
    "SnapshotUnavailable",
    "InvalidUUID",
    "InvalidAPIVersionString",
    "VersionNotFoundForAPIMethod",
    "InvalidGlobalAPIVersion",
    "ValidationError",
    "APIException",
    "APITimeout",
    "RPCTimeout",
    "Duplicate",
    "NotFound",
    "GlanceStoreNotFound",
    "GlanceStoreReadOnly",
    "VolumeNotFound",
    "MessageNotFound",
    "VolumeAttachmentNotFound",
    "VolumeMetadataNotFound",
    "InvalidVolumeMetadata",
    "InvalidVolumeMetadataSize",
    "SnapshotMetadataNotFound",
    "VolumeTypeNotFound",
    "VolumeTypeNotFoundByName",
    "VolumeTypeAccessNotFound",
    "VolumeTypeExtraSpecsNotFound",
    "VolumeTypeInUse",
    "VolumeTypeDeletionError",
    "VolumeTypeDefaultDeletionError",
    "VolumeTypeDefaultMisconfiguredError",
    "VolumeTypeProjectDefaultNotFound",
    "GroupTypeNotFound",
    "GroupTypeNotFoundByName",
    "GroupTypeAccessNotFound",
    "GroupTypeSpecsNotFound",
    "GroupTypeInUse",
    "SnapshotNotFound",
    "ServerNotFound",
    "VolumeSnapshotNotFound",
    "VolumeIsBusy",
    "SnapshotIsBusy",
    "InvalidImageRef",
    "InvalidSignatureImage",
    "ImageSignatureVerificationException",
    "ImageNotFound",
    "ServiceNotFound",
    "ServiceTooOld",
    "WorkerNotFound",
    "WorkerExists",
    "CleanableInUse",
    "ClusterNotFound",
    "ClusterHasHosts",
    "ClusterExists",
    "HostNotFound",
    "SchedulerHostFilterNotFound",
    "SchedulerHostWeigherNotFound",
    "InvalidReservationExpiration",
    "InvalidQuotaValue",
    "QuotaNotFound",
    "QuotaResourceUnknown",
    "ProjectQuotaNotFound",
    "QuotaClassNotFound",
    "QuotaUsageNotFound",
    "OverQuota",
    "FileNotFound",
    "VolumeTypeExists",
    "VolumeTypeAccessExists",
    "VolumeTypeEncryptionExists",
    "VolumeTypeEncryptionNotFound",
    "GroupTypeExists",
    "GroupTypeAccessExists",
    "GroupVolumeTypeMappingExists",
    "MalformedRequestBody",
    "ConfigNotFound",
    "ParameterNotFound",
    "NoValidBackend",
    "QuotaError",
    "VolumeSizeExceedsAvailableQuota",
    "VolumeSizeExceedsLimit",
    "VolumeBackupSizeExceedsAvailableQuota",
    "VolumeLimitExceeded",
    "SnapshotLimitExceeded",
    "UnexpectedOverQuota",
    "BackupLimitExceeded",
    "ImageLimitExceeded",
    "VolumeTypeCreateFailed",
    "VolumeTypeUpdateFailed",
    "GroupTypeCreateFailed",
    "GroupTypeUpdateFailed",
    "GroupLimitExceeded",
    "UnknownCmd",
    "MalformedResponse",
    "FailedCmdWithDump",
    "InvalidConnectorException",
    "GlanceMetadataExists",
    "GlanceMetadataNotFound",
    "ImageDownloadFailed",
    "ExportFailure",
    "RemoveExportException",
    "MetadataUpdateFailure",
    "MetadataCopyFailure",
    "InvalidMetadataType",
    "ImageCopyFailure",
    "BackupInvalidCephArgs",
    "BackupOperationError",
    "BackupMetadataUnsupportedVersion",
    "BackupMetadataNotFound",
    "VolumeMetadataBackupExists",
    "BackupRBDOperationFailed",
    "EncryptedBackupOperationFailed",
    "BackupNotFound",
    "InvalidBackup",
    "SwiftConnectionFailed",
    "TransferNotFound",
    "VolumeMigrationFailed",
    "SSHInjectionThreat",
    "QoSSpecsExists",
    "QoSSpecsCreateFailed",
    "QoSSpecsUpdateFailed",
    "QoSSpecsNotFound",
    "QoSSpecsAssociateFailed",
    "QoSSpecsDisassociateFailed",
    "QoSSpecsKeyNotFound",
    "InvalidQoSSpecs",
    "QoSSpecsInUse",
    "KeyManagerError",
    "ManageExistingInvalidReference",
    "ManageExistingAlreadyManaged",
    "InvalidReplicationTarget",
    "UnableToFailOver",
    "ReplicationError",
    "ReplicationGroupError",
    "ManageExistingVolumeTypeMismatch",
    "ExtendVolumeError",
    "EvaluatorParseException",
    "LockCreationFailed",
    "CappedVersionUnknown",
    "VolumeGroupNotFound",
    "VolumeGroupCreationFailed",
    "VolumeNotDeactivated",
    "VolumeDeviceNotFound",
    "RemoteFSException",
    "RemoteFSConcurrentRequest",
    "RemoteFSNoSharesMounted",
    "RemoteFSNoSuitableShareFound",
    "RemoteFSInvalidBackingFile",
    "NfsException",
    "NfsNoSharesMounted",
    "NfsNoSuitableShareFound",
    "ZoneManagerException",
    "FCZoneDriverException",
    "FCSanLookupServiceException",
    "ZoneManagerNotInitialized",
    "ConsistencyGroupNotFound",
    "InvalidConsistencyGroup",
    "GroupNotFound",
    "InvalidGroup",
    "InvalidGroupStatus",
    "CgSnapshotNotFound",
    "InvalidCgSnapshot",
    "GroupSnapshotNotFound",
    "InvalidGroupSnapshot",
    "InvalidGroupSnapshotStatus",
    "ISCSITargetCreateFailed",
    "ISCSITargetRemoveFailed",
    "ISCSITargetAttachFailed",
    "ISCSITargetDetachFailed",
    "TargetUpdateFailed",
    "ISCSITargetHelperCommandFailed",
    "BadHTTPResponseStatus",
    "BadResetResourceStatus",
    "MetadataAbsent",
    "NotSupportedOperation",
    "AttachmentSpecsNotFound",
    "InvalidName",
    "ServiceUserTokenNoAuth",
    "RekeyNotSupported",
    "ImageCompressionNotAllowed",
    "CinderAcceleratorError",
    "SnapshotLimitReached",
]
